import signal, logging
import importlib, time, os, inspect
import json, mimetypes
from typing import Optional, Awaitable
from concurrent.futures import ThreadPoolExecutor

import config
import tornado
from tornado import web
from tornado.concurrent import run_on_executor
from .exception import *
from .contain.controller import Controller
from .settings import SETTINGS

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class ProcessorController(web.RequestHandler):

    executor = ThreadPoolExecutor(max_workers=config.Config.web_thr_count(SETTINGS['pro_thr_num']))

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        self.request.method = 'GET'
        resp_content = yield self._dispose(*args)
        if resp_content is True:
            self.finish()
            return
        if self.response_write(resp_content) is False:
            self.finish()
            return

    @tornado.gen.coroutine
    def post(self, *args, **kwargs):
        self.request.method = 'POST'
        resp_content = yield self._dispose(*args)
        if resp_content is True:
            # self.finish()
            return
        if self.response_write(resp_content) is False:
            self.finish()
            return

    def return_file(self, path):
        """
        如果是文件且存在就处理
        PS: http://127.0.0.1:8107/static/assets/img/logo.png
        """
        file_path = "{}{}".format(SETTINGS['static_path'], path)
        if os.path.exists(file_path) and os.path.isfile(file_path):
            mime = mimetypes.guess_type(file_path)
            self.set_header('Content-Type', mime[0])
            self.set_header('cache-control', "max-age={}".format(SETTINGS.get('resource_max_age', 86400)))
            # 开启此选项浏览器回直接下载文件
            # self.set_header('Content-Disposition', 'attachment; filename=' + os.path.basename(file_path))
            buf_size = 1024 * 1024 * 10
            with open(file_path, 'rb') as f:
                while True:
                    data = f.read(buf_size)
                    if not data:
                        break
                    self.write(data)

            return True

        return False

    @run_on_executor
    def _dispose(self, *args):
        try:
            # 初始数据
            self._on_mtime = int(time.time() * 1000)
            self._is_finish = False
            # 请求信息
            request = self.request
            request.method = 'POST' if hasattr(self, 'method') is None else request.method
            logger.debug("method={}, args={}".format(request.method, args))
            remote_ip = request.remote_ip
            path = self.path = self.request.path
            path = path.replace("//", "/")
            path_arr = self.path_arr = [] if path == '/' else path.split('/')
            # 是否是ajax请求
            self.request.is_ajax = True if request.headers.get('x-requested-with') == 'XMLHttpRequest' else False
            # 是否是资源文件下载
            if self.return_file(path):
                return True
            # 收集请求参数
            params = self.params = {}
            for key in request.arguments:
                self.params[key] = self.get_argument(key)
            logger.debug("path={}, params={}".format(path, params))

            # 寻找控制器
            controller, controller_action = self.load_controller_action(self, path, path_arr)
            controller.__initialize_request__(path, params, request)

            # 执行控制器方法
            # hooker.trigger('request_before', controller)
            ret = controller_action()
            # hooker.trigger('request_after', controller)
            return "None" if ret is None else ret

        except ResponseException as e:
            return e

        except BaseException as e:
            logging.exception(e)
            return "请求异常：{}".format(e)

        finally:
            logging.info("{} {} {}ms".format(self.get_status(), self.request.path, int(time.time() * 1000) - self._on_mtime))

    MODULES = {}

    @classmethod
    def load_module(cls):
        """预加载控制器"""
        if len(cls.MODULES) > 0:
            return

        app = "app"
        c_dir_name = "controller"

        def load_method(c_class):
            methods = {}
            for it in dir(c_class):
                if it[:1] == '_': continue
                if it == 'initialize': continue
                it_cls = getattr(c_class, it)
                if inspect.isfunction(it_cls):
                    methods[it] = it

            return methods

        def load_controller(c_path, p_name=""):
            controllers = {}
            for c in cls.scan_file(c_path):
                cls_file_path = c_path + "/" + c
                if os.path.isfile(cls_file_path):
                    if c[-3:] != ".py": continue
                    cls_file_path = cls_file_path.replace('/', '.')
                    c_file = importlib.import_module(cls_file_path.replace('.py', ''))
                    # logger.debug("cls_file_path={}\nc_file={}\ndir={}".format(cls_file_path, c_file, dir(c_file)))
                    # 扫描文件里的控制器类
                    for cls_name in cls.scan_file_class(c_file):
                        c_class = getattr(c_file, cls_name)
                        if isinstance(c_class, type) is False:
                            continue
                        if c_class == Controller:
                            continue
                        if issubclass(c_class, Controller):
                            c_name = cls_name.lower().replace("controller", '')
                            if len(p_name) > 0:
                                c_name = p_name + "." + c_name
                            controllers[c_name] = {'obj': c_class(), 'm': load_method(c_class)}

                elif os.path.isdir(cls_file_path):
                    curr_name = cls_file_path.split('/')[-1]
                    if len(p_name) > 0:
                        curr_name = p_name + "." + curr_name
                    controller_subs = load_controller(cls_file_path, curr_name)
                    # logger.debug("controller_subs={}".format(controller_subs))
                    controllers = {**controllers, **controller_subs}

            return controllers

        for m in cls.scan_dir(app):
            c_path = "{}/{}/{}".format(app, m, c_dir_name)
            cls.MODULES[m] = load_controller(c_path)        # 加载模块控制器

        logger.debug("cls.MODULES={}".format(cls.MODULES))

    @classmethod
    def scan_dir(cls, path):
        ret = []
        for it in os.listdir(path):
            if it[2:] != '__':
                ret.append(it)
        ret.sort()
        return ret

    @classmethod
    def scan_file(cls, path):
        ret = []
        for it in os.listdir(path):
            if it[:1] != '_':
                ret.append(it)
        ret.sort()
        return ret

    @classmethod
    def scan_file_class(cls, file):
        ret = []
        for it in dir(file):
            if it[-2:] != '__':
                ret.append(it)
        return ret

    @classmethod
    def load_controller_action(cls, pro_obj, path, path_arr):
        """找到对应控制器和方法"""
        pa = path_arr[1:] if len(path_arr) > 0 and path_arr[0] == '' else path_arr
        if len(pa) == 0 or pa[0] == '':
            pa.extend(["index", "index", "index"])
        elif len(pa) == 1:
            pa.extend(["index", "index"])
        elif len(pa) == 2:
            pa.extend(["index"])
        else:
            pass
        logger.debug("path_arr={} -> pa={}".format(path_arr, pa))

        cls.load_module()

        module, controller, action = pa[0], '.'.join(pa[1:-1]), pa[-1]
        module, controller, action = module.lower(), controller.lower(), action.lower()
        if module in cls.MODULES:
            if controller in cls.MODULES[module]:
                if action in cls.MODULES[module][controller]['m']:
                    obj = cls.MODULES[module][controller]['obj']
                    obj = obj.__class__()
                    obj.__initialize__(pro_obj, module, controller, action)
                    return obj, getattr(obj, action)
                else:
                    raise ResponseNotFoundException("方法不存在：{}".format(action))
            else:
                raise ResponseNotFoundException("控制器不存在：{}".format(controller))
        else:
            raise ResponseNotFoundException("模块不存在：{}".format(module))

    def response_write(self, ret, status_code=200):
        """返回结果统一处理"""
        # headers
        self.set_status(status_code)
        self.set_header('Server', 'quickpython-1.05')
        self.set_header('Accept-Language', 'zh-CN,zh;q=0.9')
        self.set_header('Content-Type', 'text/html; charset={}'.format(SETTINGS['encoding']))

        if isinstance(ret, ResponseRenderException):
            return self.render(template_name=ret.tpl_name, **ret.data)
        if isinstance(ret, ResponseNotFoundException):
            status_code = 404
            ret = ret.text
        if isinstance(ret, ResponseTextException):
            ret = ret.text

        self.set_status(status_code)
        # 返回json
        request_type = self.request.headers.get("Content-Type", '').lower()
        if request_type == 'application/json' or self.request.is_ajax:
            self.set_header('Content-Type', 'application/json')
            ret = ret if isinstance(ret, Result) else Result.success(ret)
            self.write(str(ret))
            return True

        # 返回html
        if self._is_finish is False:
            self.write(str(ret))
            return True

        return False

    def on_finish(self):
        self._is_finish = True


class Result(object):

    def __init__(self, code=200, msg="SUCCESS", data=None):
        self.code = code
        self.msg = msg
        self.data = data

    def __str__(self):
        ret = {'code': self.code, 'msg': self.msg, 'data': self.data}
        return json.dumps(ret, indent=99)

    @staticmethod
    def success(data=None, msg="SUCCESS"):
        return Result.result(200, msg, data)

    @staticmethod
    def error(msg="ERROR", code=500):
        return Result.result(code, msg, None)

    @staticmethod
    def result(code, msg, data):
        return Result(code, msg, data)




"""
query_set
"""
import logging
from libs.utils import Utils
from quickpython.database.query import QuerySet, Db

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class QuerySetDemo:

    def call(self):
        # self.base()
        # self.join()
        self.trans()

    def base(self):
        self.insert()
        self.delete()
        self.update()
        self.find()

    def insert(self):
        mtime = Utils.mtime()
        QuerySet().table("user").insert({'username': "query_1"})
        QuerySet().table("user").insert({'username': "query_2"})
        QuerySet().table("user").insert({'username': "query_3"})
        QuerySet().table("user").insert({'username': "query_4"})
        QuerySet().table("user").insert({'username': "query_5"})
        logger.info("新增耗时 {}ms".format(Utils.mtime() - mtime))

    def delete(self):
        user_2 = QuerySet().table("user").where('username', 'query_2').find()
        if user_2 is None:
            raise Exception("未找到nickname = test3 的数据")
        QuerySet().table("user").where('id', user_2['id']).delete()
        logger.info("删除成功")

    def update(self):
        user_3 = QuerySet().table("user").where('username', 'query_3').find()
        if user_3 is None:
            raise Exception("未找到nickname = test3 的数据")
        else:
            QuerySet().table("user").where('id', user_3['id'])\
                .update({'nickname': user_3['username']})
            logger.info("更新成功")

    def find(self):
        user_4 = QuerySet().table("user").where('username', 'query_4').find()
        if user_4 is None:
            raise Exception("未找到nickname = test4 的数据")
        else:
            logger.info("user_4={}".format(user_4))
            logger.info("找到" + str(user_4['id']))

        # 多条
        user_rows = QuerySet().table("user").where('username', 'query_4').select()
        logger.info("user_rows={}:{}".format(len(user_rows), user_rows))

    def join(self):
        self.join_1()
        self.join_2()

    def join_1(self):
        # 查询一对一
        ret_2 = QuerySet().table("user")\
            .join('user_device', 'user_device.user_id = user.id')\
            .where('user.id', 40830)\
            .find()
        logger.info("查询一对一={}:{}".format(len(ret_2), ret_2))
        Utils.print_dicts(ret_2)

    def join_2(self):
        # 查询一对多
        ret_2 = QuerySet().table("user")\
            .join('user_money_log', 'user_money_log.user_id = user.id', 'LEFT')\
            .where('user.id', 40830)\
            .find()
        logger.info("查询一对多={}:{}".format(len(ret_2), ret_2))
        Utils.print_dicts(ret_2)

    def trans(self):
        """测试事务操作"""
        Db.start_trans()
        try:
            Db().table("user").where('id', 40830).find()
            Db().table("user").where('id', 40830).update({'xx_token': Utils.datetime_now()})
            raise Exception("test")
            Db.commit()

        except BaseException as e:
            Db.rollback()
            logger.error("出现异常", e)

    def extend(self):
        for it in range(5):
            ret = QuerySet().__get_column__(list, "user")
            logger.debug("ret={}".format(ret))


if __name__ == '__main__':
    try:
        QuerySetDemo().call()
    except BaseException as e:
        Utils.print_exc()



"""
初始化工作
"""
import logging


class DatabaseCore:

    def __init__(self):
        self.init_logging()

    def init_logging(self):
        """初始化日志"""
        formatter_str = '%(asctime)s %(levelname)s [%(filename)s:%(funcName)s:%(lineno)d]\t%(message)s'
        logging.basicConfig(format=formatter_str, level=logging.DEBUG)


databaseCore = DatabaseCore()


"""
    定义模型数据类型
"""
import time

from ..contain.func import *
from .. import setttings

options = setttings.DATABASES['default']['options']


class ColumnCmp:
    """
    字段比较描述类
    """
    def __init__(self, column, eq, val):
        self.column = column
        self.eq = eq
        self.val = val

    def __str__(self):
        return str(self.__dict__())

    def __dict__(self):
        ret = {'key': self.column.name, 'type': '=', 'val': self.val}
        if self.eq == 'gt': ret['type'] = '>'
        if self.eq == 'ge': ret['type'] = '>='
        if self.eq == 'lt': ret['type'] = '<'
        if self.eq == 'le': ret['type'] = '<='
        if self.eq == 'eq': ret['type'] = '='
        if self.eq == 'ne': ret['type'] = '!='
        return ret


class ColumnBase:
    type_ = None

    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop('name', None)       # Field:字段名称
        self.primary_key = primary_key = kwargs.pop("primary_key", False)

        self.nullable = kwargs.pop('nullable', True)      # 是否为空，不为空将开启入库检测
        if primary_key:
            self.nullable = False       # 主键禁止为空

        self.default = kwargs.pop("default", None)
        self.insert_default = kwargs.pop("insert_default", None)
        self.update_default = kwargs.pop("update_default", None)

        self.comment = kwargs.pop("comment", None)  # 字段备注
        self.proxies = kwargs.pop("proxies", None)  # 字段前缀

        self.value = kwargs.pop("value", self.default)   # 字段数据

    def has_insert_default(self):
        return (
            self.default is not None
            or self.insert_default is not None
            or self.update_default is not None
        )

    def has_update_default(self):
        return (
            self.default is not None
            or self.update_default is not None
        )

    def __get_insert_default__(self):
        val = self.default
        if not_empty(val):
            return val
        if not_empty(self.insert_default):
            if isfunction(self.insert_default):
                val = self.insert_default()
            else:
                val = self.insert_default

        if empty(val) and not_empty(self.update_default):
            if isfunction(self.update_default):
                val = self.update_default()
            else:
                val = self.update_default
        return val

    def __get_update_default__(self):
        val = self.default
        if empty(val) and not_empty(self.update_default):
            if isfunction(self.update_default):
                val = self.update_default()
            else:
                val = self.update_default
        return val

    def __gt__(self, other):    # 大于：>
        return ColumnCmp(self, 'gt', other)

    def __ge__(self, other):    # 大于等于：>=
        return ColumnCmp(self, 'ge', other)

    def __lt__(self, other):    # 小于：<
        return ColumnCmp(self, 'lt', other)

    def __le__(self, other):    # 小于等于：<=
        return ColumnCmp(self, 'le', other)

    def __eq__(self, other):    # 相等：==
        return ColumnCmp(self, 'eq', other)

    def __ne__(self, other):    # 不等：!=
        return ColumnCmp(self, 'ne', other)

    # def __str__(self):
    #     return self.value if isinstance(self.value, str) else str(self.value)
    #
    # def __int__(self):
    #     return self.value if isinstance(self.value, int) else int(self.value)


class Int(ColumnBase):
    type_ = 'int'


class Tinyint(ColumnBase):
    type_ = 'tinyint'


class Bigint(ColumnBase):
    type_ = 'bigint'


class Float(ColumnBase):
    type_ = 'float'


class Decimal(ColumnBase):
    type_ = 'decimal'


class Char(ColumnBase):
    type_ = 'char'


class Varchar(ColumnBase):
    type_ = 'varchar'

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self.length = kwargs.pop("length")


class Text(ColumnBase):
    type_ = 'text'


class Datetime(ColumnBase):
    type_ = 'datetime'


class Date(ColumnBase):
    type_ = 'date'


class Time(ColumnBase):
    type_ = 'time'


class Enum(ColumnBase):
    type_ = 'enum'


field_map = {
    Int.type_: Int,
    Tinyint.type_: Tinyint,
    Bigint.type_: Bigint,
    Float.type_: Float,
    Decimal.type_: Decimal,
    Char.type_: Char,
    Varchar.type_: Varchar,
    Text.type_: Text,
    Datetime.type_: Datetime,
    Date.type_: Date,
    Time.type_: Time,
}


"""
    字段扩展方法
"""


class ColumnFunc:

    @staticmethod
    def now():
        import datetime
        fmt = "%Y-%m-%d %H:%M:%S"
        if hasattr(options, 'datetime_fmt') and not_empty(options['datetime_fmt']):
            fmt = options['datetime_fmt']
        return datetime.datetime.now().strftime(fmt)

    @staticmethod
    def now_timestamp():
        return int(time.time())


def iscolumn(cls):
    return issubclass(cls, ColumnBase)


"""
    模型忽略属性名称
    模型处理时，直接返回值
    暂时不用，只要属性带下划线_就是自带属性，直接返回
"""
ModelAttr = [
    "__class__",
    "__modified_fields__",
    "__table__",
    "__withs__",
    "__query__",
    "__attrs__",
]



"""
    助手方法
"""
import importlib
import types
from inspect import isfunction


def empty(obj):
    if obj is None:
        return True
    elif isinstance(obj, int) and obj == 0:
        return True
    elif isinstance(obj, str) and len(obj) == 0:
        return True
    elif isinstance(obj, list) and len(obj) == 0:
        return True
    elif isinstance(obj, dict) and len(obj) == 0:
        return True
    return False


def not_empty(obj):
    return empty(obj) is False


def load_cls(cls_path, **kwargs):
    model_path, cls_name = cls_path.rsplit(".", 1)
    module = importlib.import_module(model_path)
    cls = getattr(module, cls_name)
    return cls




"""
    数据库连接管理器
"""
import pymysql
import time, threading, logging
from .. import setttings
from ..contain.func import *


class Connection:

    _local = threading.local()
    log = logging.getLogger(__name__)

    def __init__(self):
        self._is_trans_ing = False
        self._conn = None           # type:pymysql.Connection
        self._conn_cursor = None
        self._conn_start_time = int(time.time())
        self.config = {
            'engine': "mysql",
            'hostname': "127.0.0.1",
            'port': 3306,
            'database': None,
            'username': None,
            'password': None,
            'charset': "utf8mb4",
            'wait_timeout': 3600,
            'prefix': "",       # 表前缀
            'options': {},
        }
        self.config = {**self.config, **setttings.DATABASES['default']}
        self.config['port'] = int(self.config['port']) if not_empty(self.config['port']) else 3306
        self.config['wait_timeout'] = int(self.config['wait_timeout'])
        # self.log.setLevel(logging.DEBUG if self.get_config('echo') else logging.INFO)

    def get_config(self, name, def_val=None):
        if name in self.config:
            return self.config[name]
        return self.config['options'][name] if name in self.config['options'] else def_val

    @staticmethod
    def get_connect():
        if hasattr(Connection._local, '__db_connection'):
            return Connection._local.__getattribute__('__db_connection')

        connection = Connection()
        connection.connect()
        Connection._local.__setattr__('__db_connection', connection)
        return connection

    def connect(self, re=False):
        """连接 已连接将返回"""
        if self._conn is None or re:
            self.log.debug("connect={}".format({'hostname': self.config['hostname']}))
            if self.config['engine'] == 'mysql':
                self._conn = pymysql.connect(
                    host=self.config['hostname'],
                    port=self.config['port'],
                    user=self.config['username'],
                    password=self.config['password'],
                    db=self.config['database'],
                    charset=self.config['charset'])
                self._conn.autocommit(True)

            else:
                raise Exception("不支持的数据库引擎：{}".format(self.config['engine']))
            self._conn_start_time = int(time.time())

        return self

    def check_wait_timeout(self):
        """检测超时，超时将重新连接（处于事务中时不会重连）"""
        if self._is_trans_ing:      # 事务中不检测
            return False
        curr_time = int(time.time())
        if curr_time - self._conn_start_time >= self.config['wait_timeout']:
            self._conn_cursor.close()
            self._conn_cursor = None
            self._conn.close()
            self._conn = None
            self.connect(True)

    def ping(self):
        """数据库PING"""

    def start_trans(self):
        """开启事务"""
        self.check_wait_timeout()       # TODO::获取游标时检测超时
        self.log.debug("start_trans")
        self._conn.autocommit(False)

    def commit(self):
        """提交事务"""
        self.log.debug("commit")
        self._conn.commit()

    def rollback(self):
        """回滚事务"""
        self.log.debug("rollback")
        self._conn.rollback()
        self._conn.autocommit(True)

    def get_cursor(self):
        self.check_wait_timeout()       # TODO::获取游标时检测超时
        if self._conn_cursor is None:
            self._conn_cursor = self._conn.cursor()
        return self._conn_cursor

    def execute(self, sql):
        self.log.debug(sql)
        cur = self.get_cursor()
        count = cur.execute(sql)
        ret = cur.fetchone()
        return count, ret, cur.description

    def execute_all(self, sql):
        self.log.debug(sql)
        cur = self.get_cursor()
        count = cur.execute(sql)
        ret = cur.fetchall()
        return count, ret, cur.description

    def execute_get_id(self, sql):
        self.log.debug(sql)
        cur = self.get_cursor()
        count = cur.execute(sql)
        ret = cur.fetchone()
        ret_id = self._conn.insert_id()
        return count, ret, ret_id

    def close(self):
        self._conn.close()



import quickpython.component.function as functions
from quickpython.config import Config, env


# app配置
SETTINGS = {
    'debug': env.get('app.debug', False),
    'port': env.get('app.port', False),
    'template_path': 'app',
    'static_path': 'public',
    'ui_methods': functions,
    'pro_thr_num': None,        # 默认是CPU核心数的线程数量
    # 环境配置
    'lang': 'zh-cn',
    'timezone': 'PRC',
    'encoding': 'utf-8',        # 全局框架文件默认编码
    # 应用设置
    'default_module': 'index',
    'default_controller': 'index',
    'default_action': 'index',
    'url_convert': True,    # 自动转换为控制器方法
    'url_route_on': False,
    # 模板配置
    'template': {
        'view_path': '',
        'view_suffix': 'html',
        'cache': True,
    },
    'view_replace_str': {
        '__PUBLIC__': '',
        '__ROOT__': '',
        '__CDN__': '',
    },
    'dispatch_success_tmpl': "",
    'dispatch_error_tmpl': "",
    'dispatch_exception_tmpl': "",
    'error_message': "你所浏览的页面暂时无法访问",
    'exception_handle': '',
    # 验证码
    'captcha': {
        'code_set': '2345678abcdefhijkmnpqrstuvwxyzABCDEFGHJKLMNPQRTUVWXY',
        'font_size': 18,
        'height': 40,
        'width': 130,
        'count': 5,
    },
    #
    'resource_max_age': 86400,
}

# session配置
SESSIONS = {
    'prefix': 'qp_',
    'type': "",
    'expire': 86400 * 7,
}

# 日志配置
LOGGING = {
    'default': {
        'level': 'DEBUG',
    }
}

# 路由
from .processor import ProcessorController
ROUTES = [
    (r"/(.*)", ProcessorController),    # 默认处理控制器
]



"""
    查询基础操作集合
    环境：pymysql>=1.0.2
"""
import logging, time, re, pymysql
from .connector.connection import Connection
from .contain.func import *
from .contain.component import *
from . import setttings

logger = logging.getLogger(__name__)

options = setttings.DATABASES['default']['options']


def format_field(name):
    if options['field_sep'] is None or options['field_sep'] is True:
        arr = name.split(".")
        return "`{}`".format(name) if len(arr) == 1 else "`{}`.`{}`".format(arr[0], arr[1])
    return name


def format_val(val):
    if val is None:
        ret = 'NULL'
    elif isinstance(val, int):
        ret = str(val)
    elif isinstance(val, float):
        ret = str(val)
    else:
        val = pymysql.converters.escape_string(str(val))        # 特殊符号转义，防止SQL注入
        ret = '{}{}{}'.format(options['val_str_sep'], val, options['val_str_sep'])

    return ret


def typeof(variate):
    mold = None
    if isinstance(variate, int):
        mold = "int"
    elif isinstance(variate, str):
        mold = "str"
    elif isinstance(variate, float):
        mold = "float"
    elif isinstance(variate, list):
        mold = "list"
    elif isinstance(variate, tuple):
        mold = "tuple"
    elif isinstance(variate, dict):
        mold = "dict"
    elif isinstance(variate, set):
        mold = "set"
    return mold


class QuerySet(object):

    def __init__(self):
        self.__map = {}
        self.__database = self.__conn.config['database']
        self.__table = ''
        self.__fields = []
        self.__where = {}
        self.__alias = ''
        self.__join = []
        self.__having = []
        self.__distinct = False
        self.__model__ = None
        self.__options = {
            'multi': {},
            'where': {},
        }
        self.prefix = self.__conn.get_config("prefix", '')

    @property
    def __conn(self):
        return Connection.get_connect()

    def __close(self):
        self.__conn.close()

    @property
    def __self(self):
        return self if self.__model__ is None else self.__model__

    @property
    def __table__(self):
        """完整表名（带前缀）"""
        return self.__table if self.prefix is None else self.prefix + self.__table

    def table(self, table):
        self.__table = table
        return self

    def where(self, field, op=None, condition=None):
        self.__parse_where_exp("AND", field, op, condition)
        return self

    def where_raw(self, field):
        self.__parse_where_exp("AND", field, op=None, condition=None)
        return self

    def __parse_where_exp(self, logic, field, op, condition, param=None, strict=False):
        logic = logic.upper()
        param = [] if param is None else param
        where = {}
        # logic
        if logic not in self.__map:
            self.__map[logic] = {}

        # 匿名方法
        if isfunction(field):
            field(self)
        # 自写语句
        elif isinstance(field, str) and len(re.findall('[,=\>\<\'\"\(\s]', field)) > 0:
            field = Expression(field)
            self.__map[logic][str(field)] = ['exp', field]
        # 条件组装
        elif op is None and condition is None:
            if isinstance(field, dict):
                where = field
                for k, val in field.items():
                    if isinstance(val, list):
                        self.__map[logic][k] = val
                    else:
                        self.__map[logic][k] = ['=', val]

            elif isinstance(field, str):
                where[field] = ['null', '']
                self.__map[logic][field] = where[field]

        elif empty(condition):
            where[field] = ['=', op]
            self.__map[logic][field] = where[field]
        else:
            if op == 'exp':
                where[field] = ['exp', Expression(condition)]
            else:
                where[field] = [op, condition]
            self.__map[logic][field] = where[field]

        if len(where) > 0:
            if logic not in self.__options['where']:
                self.__options['where'][logic] = {}

    def limit(self, start=0, limit=None):
        if limit is None:
            limit = start
            start = 0
        self.__options['limit'] = "{}, {}".format(start, limit)
        return self

    def page(self, page=1, page_size=20):
        self.limit((page - 1) * page_size, page_size)
        return self

    def order(self, field, asc='ASC'):
        if 'order' not in self.__options:
            self.__options['order'] = []
        self.__options['order'].append("{} {}".format(format_field(field), asc))
        return self

    def group(self, group):
        self.__options['group'] = group
        return self

    def field(self, field, is_except=False, table_name=None, prefix=None, alias=None):
        if empty(field):
            return self

        def array_diff(arr, arr2):
            return arr      # 用的排除

        if field is True:
            fields = self.__get_column__(list, self.__table__ if table_name is None else table_name)
            field = ['*'] if fields is None else fields
        elif is_except:
            fields = self.__get_column__(list, self.__table__ if table_name is None else table_name)
            field = field if fields is None else array_diff(fields, field)

        if not_empty(table_name):
            prefix = table_name if prefix is None else prefix
            for idx in range(len(field)):
                val = field[idx]
                val = "{}.{}{}".format(prefix, val, ('' if alias is None else " AS {}{}".format(alias, val)))
                field[idx] = val

        self.__fields.extend(field)
        return self

    def distinct(self, is_true=True):
        self.__distinct = is_true
        return self

    def having(self, where):
        self.__having.append(where)
        return self

    __op_map = {
        'eq': '=',
        'neq': '<>',
        'lt': '<',
        'elt': '<=',
        'gt': '>',
        'egt': '>=',
    }

    __op_in = {
        'in': 'IN',
        'not in': 'NOT IN',
        'exists': 'EXISTS',
    }

    def get_map(self):
        return self.__map

    def __com_where_sql(self, to_str=True):
        """构建WHERE条件"""
        # logger.debug("self.__map={}".format(self.__map))
        sa = []
        for logic, items in self.__map.items():
            if len(items) == 0:
                continue
            sa_where = []
            for field, condition in items.items():
                # logger.debug("condition={}".format(condition))
                conditions = condition if isinstance(condition[0], list) else [condition]
                for cond in conditions:
                    op = str(cond[0]).lower()
                    if op == 'exp':
                        where_item = str(cond[1])      # 表达式查询，支持SQL语法
                    elif op in self.__op_map:
                        where_item = "{}{}{}".format(format_field(field), self.__op_map[op], format_val(cond[1]))
                    elif op in self.__op_in:
                        in_val = cond[1]
                        if isinstance(cond[1], list):
                            in_val = [format_val(it) for it in cond[1]]
                            in_val = ",".join(in_val)
                        where_item = "{}{}({})".format(format_field(field), self.__op_in[op], in_val)
                    else:
                        where_item = "{}{}{}".format(format_field(field), cond[0], format_val(cond[1]))

                    sa_where.append(where_item)

            if len(sa) > 0:
                sa.append(logic)  # 前面有条件了，追加本次的条件
            if len(sa_where) > 0:
                sa.append(" {} ".format(logic).join(sa_where))

        if len(sa) == 0:
            return ""
        sa.insert(0, "WHERE")
        return " ".join(sa) if to_str else sa

    def __com_query_sql(self):
        """公共查询SQL"""
        if self.__table__ is None:
            return None

        sa = ["SELECT"]
        if self.__distinct:
            sa.append("distinct")

        fields = ','.join(self.__fields)
        sa.append("{}".format(fields if len(fields) > 0 else '*'))

        table_name = '' if self.prefix == '' else ' `{}`'.format(self.__table)
        sa.append("FROM {}{}".format(self.__table__, table_name))
        if self.__alias != '':
            sa.append(self.__alias)

        if len(self.__join) > 0:
            for item in self.__join:
                table, alias = item['table'], ''
                if isinstance(table, dict):
                    for key in table:
                        table, alias = key, table[key]
                table = self.__complement_table_name(table)
                sa.append("{} JOIN {} {} ON {}".format(item['type_'], table, alias, item['on']))

        sa.append(self.__com_where_sql())

        if 'group' in self.__options:
            sa.append("GROUP BY {}".format(self.__options['group']))

        if len(self.__having) > 0:
            for item in self.__having:
                sa.append(item)

        if self.__options.get('order'):
            sa.append("ORDER BY {}".format(",".join(self.__options.get('order', []))))

        if self.__options.get('limit'):
            sa.append("LIMIT {}".format(self.__options['limit']))
        return " ".join(sa)

    def build_sql(self):
        column = self.__fields

        if column == '*':
            column = self.__get_column__(str)

        sql = self.__com_query_sql()
        return sql

    def __get_field(self, table=None):
        column = self.__fields
        if '*' in column:
            column = self.__get_column__(str, table=table)

        return column

    """
    数据执行操作（调用即执行）
    """

    def find(self):
        sql = self.__com_query_sql()
        if sql is None or len(sql) == 0:
            return None

        # sql += " LIMIT 1"
        # logger.debug("{} {}".format(id(self.__conn), sql))
        count, result, field_info = self.__conn.execute_all(sql)
        if result is None or len(result) == 0:
            return None

        # 根据返回字段进行数据封装
        data = {}
        row = result[0]
        for idx in range(len(field_info)):
            name = field_info[idx][0]
            data[name] = row[idx]

        return data

    def select(self):
        sql = self.__com_query_sql()
        if sql is None:
            return None

        count, result, field_info = self.__conn.execute_all(sql)
        ret = []

        for index, item in enumerate(result):
            data = {}
            for idx in range(len(field_info)):
                name = field_info[idx][0]
                data[name] = item[idx]
            ret.append(data)

        return ret

    def value(self, field):
        self.__fields = [field]
        sql = self.__com_query_sql()
        count, result, _ = self.__conn.execute(sql)
        if result is not None:
            return result[0]
        else:
            return 0

    def count(self):
        return self.value('count(*)')

    def insert(self, data):
        if isinstance(data, dict) is False:
            return None

        fields, values = [], []
        for key, val in data.items():
            fields.append(format_field(key))
            values.append(format_val(val))

        if len(fields) == 0 or len(values) == 0:
            return 0

        sql = "INSERT INTO {}({}) VALUES({})".format(self.__table__, ", ".join(fields), ", ".join(values))
        return self.__conn.execute(sql)[0]

    def insert_get_id(self, data):
        if isinstance(data, dict) is False:
            return None

        fields, values = [], []
        for key, val in data.items():
            fields.append(format_field(key))
            values.append(format_val(val))

        if len(fields) == 0 or len(values) == 0:
            return 0

        sql = "INSERT INTO {}({}) VALUES({})".format(self.__table__, ", ".join(fields), ", ".join(values))

        # 执行
        count, ret, pk = self.__conn.execute_get_id(sql)
        # logger.debug("pk={}".format(pk))
        return pk

    def insert_all(self, datas):
        if typeof(datas) != 'list':
            return None
        count = 0
        for data in datas:
            count += self.insert(data)
        return count

    def update(self, data):
        if isinstance(data, dict) is False:
            return None

        if len(self.__map) == 0:
            raise Exception("禁止不使用 where 更新数据")
        sql_where = self.__com_where_sql()

        fields = []
        for key, val in data.items():
            fields.append("{}={}".format(format_field(key), format_val(val)))

        if len(fields) == 0:
            return 0

        # 表名、更新的字段、限制条件
        sql = "UPDATE {} SET {} {}".format(format_field(self.__table__), ", ".join(fields), sql_where)
        count, ret, _ = self.__conn.execute(sql)
        return count

    def set_option(self, key, val):
        return self.update({key: val})

    def delete(self):
        sql_where = self.__com_where_sql()
        if len(sql_where.strip()) == 0:
            raise Exception("禁止不使用 where 删除数据")

        sql = "DELETE FROM {} {}".format(self.__table__, sql_where)
        count, result, _ = self.__conn.execute(sql)
        return count

    def set_inc(self, key, step=1):
        return self.update({key: key + '+' + str(step)})

    def set_dec(self, key, step=1):
        return self.update({key: key + '-' + str(step)})

    def query(self, sql):
        count, ret, _ = self.__conn.execute(sql)
        return ret

    @staticmethod
    def start_trans():
        """开启事务"""
        Connection.get_connect().start_trans()

    @staticmethod
    def commit():
        """提交事务"""
        Connection.get_connect().commit()

    @staticmethod
    def rollback():
        """回滚事务"""
        Connection.get_connect().rollback()

    def __complement_table_name(self, name:str):
        pre_len = len(self.prefix)
        if pre_len > 0:
            if name[:pre_len] != self.prefix:
                name = self.prefix + name

        return name

    def __get_column__(self, conv=None, table=None, full_name=False):
        """获取表字段"""
        table = self.__table__ if table is None else self.__complement_table_name(table)
        ck = "{}.{}_{}".format(self.__database, table, full_name)
        cache_data = self.__get_cache(ck)
        if cache_data is not None:
            count, list_data, field_info = cache_data
        else:
            sql = "SHOW FULL COLUMNS FROM `{}`".format(table)
            count, list_data, field_info = self.__conn.execute_all(sql)
            self.__set_cache(ck, (count, list_data, field_info))

        if full_name:
            all_column = list([{'field': "{}.{} AS {}__{}".format(table, item[0], table, item[0]),
                                'field_as': "{}__{}".format(table, item[0]),
                                'field_': item[0],
                                'type': item[1], 'key': item[4]} for item in list_data])
        else:
            all_column = list([{'field': item[0],
                                'field_as': item[0],
                                'field_': item[0],
                                'type': item[1], 'key': item[4]} for item in list_data])

        if conv is str:
            return ','.join([it['field'] for it in all_column])
        if conv is list:
            return [it['field'] for it in all_column]

        return all_column

    def __get_pk__(self):
        """获取表的主键"""
        fields = self.__get_column__()
        for field in fields:
            if field["key"] == 'PRI':
                return field['field']
        return None

    """
    join查询
    """
    def join(self, table, on, type_="INNER"):
        """
        usage:  query = QuerySet().table('user')
            1.  query.join('user_info', 'user_info.user_id = user.id', 'LEFT')
            2.  query.join('user_info', 'user_info.user_id = user.id')
        """
        info = {'table': table, 'on': on, 'type_': type_}
        self.__join.append(info)
        return self


    """
    数据缓存
    """
    __column_cache__ = {}

    def __get_cache(self, ck, default_value=None):
        item = self.__column_cache__.get(ck, default_value)
        if item is None:
            return None
        elif item['end_time'] <= int(time.time()):
            return None
        return item['data']

    def __set_cache(self, ck, value, timeout=60):
        item = {
            'end_time': int(time.time()) + timeout,
            'data': value,
        }
        self.__column_cache__[ck] = item
        return item['data']


class Db(QuerySet):
    """"""



"""
    ORM框架配置
"""
from quickpython.component.env import env


# 数据库连接配置
DATABASES = {
    'default': {
        'engine': 'mysql',
        'hostname': env.get('database.hostname', "127.0.0.1"),
        'database': env.get('database.database', "root"),
        'username': env.get('database.username', "root"),
        'password': env.get('database.password', "root"),
        'port': env.get('database.port', 3306),
        'prefix': env.get('database.prefix', ""),
        'echo': env.get('database.echo', False),
        'wait_timeout': 60,
        'options': {
            'init_command': 'SET default_storage_engine=INNODB;',
            'charset': 'utf8',
            'table_prefix': "",  # 表名前缀，todo::暂不支持
            'field_sep': True,      # 开启字段表名分割符，例：`id` or id
            'val_str_sep': "\"",    # 字符数据包围符号 只能是'和"
            'with_seq': "__",       # 预载入表字段关联前缀
            'datetime_fmt': "",     # 时间格式
        },
    },
}




"""
    python-quick-captcha 1.0
"""
import logging, random, os
from PIL import Image, ImageDraw, ImageFont, ImageFilter


class Captcha:

    # @staticmethod
    # def out(request):
    #     from django.shortcuts import HttpResponse
    #     event = request.GET.get("event", 'default')
    #     image_data, code = Captcha.generate()
    #     request.session['_captcha_' + event] = code
    #     return HttpResponse(image_data, content_type="image/png")

    @staticmethod
    def verify(request, text=None, event='default'):
        true_val = request.session.get('_captcha_' + event)
        if text is None or event is None or true_val is None:
            return False
        request.session['_captcha_' + event] = None       # 每次验证后都要求刷新验证码
        return str(text).lower() == str(true_val).lower()

    @staticmethod
    def generate():
        # 绘制干扰线
        def gene_line(draw, width, height, linecolor):
            # random.randint(a, b)用于生成一个指定范围内的证书，其中第一个参数a是上限，第二个参数b是下限，生成的随机数n：a<=n<=b
            begin = (random.randint(0, width), random.randint(0, height))
            end = (random.randint(0, width), random.randint(0, height))
            # 在图像上画线，参数值为线的起始和终止位置坐标[(x, y), (x, y)]和线的填充颜色
            draw.line([begin, end], fill=linecolor, width=2)

        # 生成验证码
        def gene_code(size, font_count, font_color, line_color):
            font_path = '{}/font.ttf'.format(os.path.dirname(os.path.abspath(__file__)))
            # 宽和高
            width, height = size
            # 创建图片, 'RGBA'表示4*8位像素，真彩+透明通道
            image = Image.new('RGB', (width, height), (255, 255, 255))
            # 验证码的字体。ImageFont这个函数从指定的文件加载了一个字体对象，并且为指定大小的字体创建了字体对象。
            font = ImageFont.truetype(font_path, 30)
            # 创建画笔，创建可用于绘制给定图像的对象
            draw = ImageDraw.Draw(image)
            # 随机生成想要的字符串
            font_count = random.randint(4, font_count)
            text = random.sample('ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789', font_count)
            text = ''.join(text)
            # 返回给定文本的宽度和高度，返回值为2元组
            font_width, font_height = font.getsize(text)
            # 填充字符串,参数分别是：文本的左上角坐标，文本内容，字体，文本的填充颜色
            draw.text(((width-font_width) / font_count, (height-font_height) / font_count), text, font=font, fill=font_color)
            # 画线的条数
            line_count = random.randint(2, 4)
            for i in range(line_count):
                gene_line(draw, width, height, line_color)
            # 滤镜，边界加强，ImageFilter.EDGE_ENHANCE_MORE为深度边缘增强滤波，会使得图像中边缘部分更加明显。
            image = image.filter(ImageFilter.EDGE_ENHANCE_MORE)
            # 保存验证码图片
            # image.save('idencode.png')

            from io import BytesIO
            with BytesIO() as f:
                image.save(f, format='JPEG')
                return f.getvalue(), text

        # TODO::调用生成
        size = (120, 50)        # 生成验证码图片的高度和宽度
        font_color = (0, 0, 255)     # 字体颜色，默认为蓝色
        line_color = (100, 100, 255)     # 线颜色，默认为蓝色
        return gene_code(size, 5, font_color, line_color)


import pymysql, logging

#   pip install pymysql

#   1、实例化类 sql = Db()
#   参数类：
#   table方法：入参：表名
#       使用方法: sql.table('test')
#   where方法：可多次调用，入参：字符串或字典(key:字段名,val:值,type:对应关系,默认为'=')或数组
#       使用方法: sql.where('id=1')
#       使用方法: sql.where({'id',1})
#       使用方法: sql.where({'id',1,'<>'})
#       使用方法: sql.where({'id','(1,2)','in'})
#       使用方法: sql.where([{'id','(1,2)','in'},{'name','%我%','like'}])
#   whereRow方法:可多次调用，入参：字段名,值,对应关系
#       使用方法: sql.whereRow('id',1)
#       使用方法: sql.whereRow('id',1,'>')
#   whereIn方法:可多次调用，入参：字段名,值(逗号拼接的字符串或元组)
#       使用方法: sql.whereIn('id','1,2')
#       使用方法: sql.whereIn('id',(1,2))
#   whereLike方法:可多次调用，入参：字段名,值
#       使用方法: sql.whereLike('name','%我%')
#   alias方法:对 table 方法的表名设置 别名
#       使用方法: sql.table('test').alias('a')
#   limit方法：查询前N条数据，入参：数量
#       使用方法: sql.limit(10)
#   order方法：排序，入参：排序字符串
#       使用方法: sql.order('id desc')
#   group方法:分组，入参：字符串
#       使用方法: sql.group('name,title')
#   field方法：查询字段,入参：字符串
#       使用方法: sql.field('id,name,title,create_time')
#   distinct方法：去重，入参：True或False，默认为True
#       使用方法: sql.distinct()
#   having方法：入参：字符串
#       使用方法: sql.having('b.id=1')
# =====================================================================================
#   操作类：
#   buildSql    返回select查询语句，出参：字符串
#       使用方法: sql.buildSql()
#   find    返回符合条件的第一条数据，出参：字典
#       使用方法: sql.find()
#   count   返回符合条件的数量：出参：Int
#       使用方法: sql.count()
#   value   入参：字段名，返回符合条件的一个字段的值，出参：值
#       使用方法: sql.value('name')
#   select  返回符合条件的多数据，出参：数组<字典>
#       使用方法: sql.select()
#   get 入参：值，识别主键返回符合条件的第一条数据，出参：字典
#       使用方法: sql.get(2)
#   insert  入参：字典,字段识别后根据键名插入数据,自动剔除表中没有的字段，出参：变更数量
#       使用方法: sql.insert({'id':1,'name':'你好'})
#   insertGetId 入参：字典,字段识别后根据键名插入数据并返回自增主键值，出参：值
#       使用方法: sql.insertGetId({'id':1,'name':'你好'})
#   insertAll   入参：数组<字典>，同 insert 的数组版，出参：变更数量
#       使用方法: sql.insertAll([{'id':1,'name':'你好'}])
#   update  入参：字典，根据条件更新数据，出参：变更数量
#       使用方法: sql.update({'name':'你好'})
#   setOption 入参：字段名，变更值，根据条件更新数据，出参：变更数量
#       使用方法: sql.setOption('name','你好')
#   delete 根据条件删除数据，出参：变更数量
#       使用方法: sql.delete()
#   setInc 入参：字段名，数字(步仅值，默认为1)；字段值自增，出参：变更数量
#       使用方法: sql.setInc('num')
#       使用方法: sql.setInc('num',2)
#   setDec 入参：字段名，数字(步仅值，默认为1)；字段值自减，出参：变更数量
#       使用方法: sql.setDec('num')
#       使用方法: sql.setDec('num',2)
#   query 入参：数据库语句
#       使用方法: sql.query('delete from id=1')

HOST = '127.0.0.1'  # 链接地址
USER = 'root'  # 数据库用户
PASS = '123456'  # 数据库密码
CHAR = 'utf8'  # 编码格式
DB = 'test'  # 数据库名称
PORT = 3306  # 数据库端口号
PREFIX = 'bs_'  # 表前缀

logger = logging.getLogger(__name__)


def format_field(val, mold):
    mold = str(mold).lower()
    if 'int' in mold:
        return str(val)
    if 'decimal' in mold:
        return str(val)
    if 'float' in mold:
        return str(val)
    if 'double' in mold:
        return str(val)
    if 'char' in mold:
        return format_val(val)
    if 'text' in mold:
        return format_val(val)
    if 'date' in mold:
        return format_val(str(val))
    if 'time' in mold:
        return format_val(str(val))
    if 'year' in mold:
        return format_val(str(val))
    if 'blob' in mold:
        return format_val(val)
    return format_val(val)


def format_val(val):
    types = typeof(val)
    if types == 'str':
        return "'%s'" % str(val).replace('\'', '"')
    if types == 'int':
        return "%s" % val


def typeof(variate):
    mold = None
    if isinstance(variate, int):
        mold = "int"
    elif isinstance(variate, str):
        mold = "str"
    elif isinstance(variate, float):
        mold = "float"
    elif isinstance(variate, list):
        mold = "list"
    elif isinstance(variate, tuple):
        mold = "tuple"
    elif isinstance(variate, dict):
        mold = "dict"
    elif isinstance(variate, set):
        mold = "set"
    return mold


class Db(object):
    __conn = None
    __map = []
    __name = ''
    __column = '*'
    __alias = ''
    __join = []
    __having = []
    __distinct = False
    __option = {}

    def __init__(self, host=HOST, username=USER, password=PASS, db=DB, charset=CHAR, port=PORT, prefix=PREFIX):
        self.host = host
        self.username = username
        self.password = password
        self.db = db
        self.charset = charset
        self.port = port
        self.prefix = prefix

    def __connect(self):
        self.__conn = pymysql.connect(host=self.host, port=self.port, user=self.username, password=self.password,
                                      db=self.db, charset=self.charset)
        self.cursor = self.__conn.cursor()

    def __close(self):
        self.cursor.close()
        self.__conn.close()

    def table(self, table):
        self.__name = self.prefix + table
        return self

    def where(self, where=None):
        if where is None:
            return self
        if typeof(where) == 'dict':
            self.__map.append({'key': where.get('key'), 'val': where.get('val'), 'type': where.get('type')})

        if typeof(where) == 'list':
            for item in where:
                self.where(item)
        if typeof(where) == 'str':
            self.__map.append(where)
        return self

    def whereRow(self, key, val, mark='='):
        return self.where({'key': key, 'val': val, 'type': mark})

    def whereIn(self, key, val):
        if typeof(val) == 'tuple':
            return self.where({'key': key, 'val': '(' + ','.join(val) + ')', 'type': 'in'})
        return self.where({'key': key, 'val': '(' + val + ')', 'type': 'in'})

    def whereLike(self, key, val):
        return self.where({'key': key, 'val': val, 'type': 'like'})

    def alias(self, name=''):
        self.__alias = name + ' '
        return self

    #   字段没处理明白，暂时禁用
    # def join(self, table, where, mold='inner'):
    # table = self.prefix + table
    # self.__join.append({'table': table, 'where': where, 'mold': mold})
    # return self
    #

    def limit(self, limit=1):
        self.__option['limit'] = format_val(limit)
        return self

    def order(self, order):
        self.__option['order'] = order
        return self

    def group(self, group):
        self.__option['group'] = group
        return self

    def field(self, field):
        self.__column = field
        return self

    def distinct(self, is_true=True):
        self.__distinct = is_true
        return self

    def having(self, where):
        self.__having.append(where)
        return self

    def __comQuerySql(self):
        if self.__name is None:
            return None

        column = self.__column
        if column == '*':
            column = self.__showColumn(True)

        sql = "SELECT "
        if self.__distinct:
            sql += ' distinct '

        sql += str(column) + " from " + str(self.__name)

        if self.__alias != '':
            sql += ' ' + self.__alias

        if len(self.__join) > 0:
            if '*' in column:
                raise Exception("使用 join 必须指定字段名")
            for item in self.__join:
                sql += item['mold'] + ' join ' + item['table'] + ' on ' + item['where'] + ' '

        if len(self.__map) > 0:
            sql += ' where 1=1 '
            for item in self.__map:
                if typeof(item) == 'str':
                    sql += ' and ( ' + item + ' ) '
                elif typeof(item) == 'dict':
                    sql += ' and ( ' + item.get('key') + ' ' + item.get('type') + ' ' + str(item.get('val')) + ' ) '

        if self.__option.get('group'):
            sql += ' group by ' + self.__option['group'] + ' '

        if len(self.__having) > 0:
            for item in self.__having:
                sql += ' ' + item

        if self.__option.get('order'):
            sql += ' order by ' + self.__option['order'] + ' '

        if self.__option.get('limit'):
            sql += ' limit ' + self.__option['limit']
        return sql

    def buildSql(self):
        column = self.__column

        if column == '*':
            column = self.__showColumn(True)

        sql = self.__comQuerySql()
        return sql

    def __getField(self, table=None):
        column = self.__column
        if '*' in column:
            column = self.__showColumn(True, table=table)

        return column

    # 接下来是数据库操作

    def find(self):

        column = self.__getField()
        sql = self.__comQuerySql()
        if sql is None:
            return None
        result = None
        try:
            self.__connect()
            self.cursor.execute(sql)
            result = self.cursor.fetchone()
            self.__close()
        except Exception as e:
            logger.exception(e)
            raise e
        if result is None:
            return None
        data = {}
        for index, k in enumerate(column.split(',')):
            data[k] = result[index]
        return data

    def select(self):
        if self.__name is None:
            return

        column = self.__getField()

        sql = self.__comQuerySql()
        if sql is None:
            return None
        result = None
        try:
            self.__connect()
            self.cursor.execute(sql)
            result = self.cursor.fetchall()
            self.__close()
        except Exception as e:
            logger.exception(e)
            raise e
        if result is None:
            return None
        data = []
        sp = column.split(',')
        for index, item in enumerate(result):
            dicts = {}
            for index2, k in enumerate(sp):
                dicts[k] = result[index][index2]
            data.append(dicts)
        return data

    def get(self, vid):
        pk = self.__getPk()
        if pk is not None:
            self.whereRow(pk, vid)
        return self.find()

    def value(self, field):
        self.__column = field
        sql = self.__comQuerySql()
        result = None
        try:
            self.__connect()
            self.cursor.execute(sql)
            result = self.cursor.fetchone()
            self.__close()
        except Exception as e:
            logger.exception(e)
            raise e
        if result is not None:
            return result[0]
        else:
            return None

    def count(self):
        return self.value('count(*)')

    def insert(self, data):
        all_column = self.__showColumn()
        fields = ''
        values = ''
        i = 0
        for key in data:
            for column in all_column:
                if column['field'] == key:
                    if i == 0:
                        fields = key
                        values = format_field(data[key], column['type'])
                    else:
                        fields += ',' + key
                        values += ',' + format_field(data[key], column['type'])
                    i += 1
        if fields == '' or values == '':
            return 0
        sql = str("INSERT INTO " + self.__name + " ( " + fields + ") values ( " + values + " ) ")
        return self.__edit(sql)

    def update(self, data):
        if typeof(data) != 'dict':
            return None
        all_column = self.__showColumn()
        fields = ''
        i = 0
        sql = ''
        if len(self.__map) > 0:
            sql += ' where 1=1 '
            for item in self.__map:
                if typeof(item) == 'str':
                    sql += ' and ( ' + item + ' ) '
                elif typeof(item) == 'dict':
                    sql += ' and ( ' + item.get('key') + ' ' + item.get('type') + ' ' + str(item.get('val')) + ' ) '
        else:
            raise Exception("禁止不使用 where 更新数据")
        for key in data:
            for column in all_column:
                if column['field'] == key:
                    if i == 0:
                        fields = key + '=' + format_field(data[key], column['type'])
                    else:
                        fields += str(',' + key + '=' + format_field(data[key], column['type']))
                    i += 1
        if fields == '':
            return 0
        sql = str("UPDATE " + self.__name + " set " + fields + ' ' + sql)
        logger.debug(sql)
        return self.__edit(sql)

    def insertGetId(self, data):
        if typeof(data) != 'dict':
            return None
        all_column = self.__showColumn()
        fields = ''
        values = ''
        i = 0
        for key in data:
            for column in all_column:
                if column['field'] == key:
                    if i == 0:
                        fields = key
                        values = format_field(data[key], column['type'])
                    else:
                        fields += ',' + key
                        values += ',' + format_field(data[key], column['type'])
                    i += 1
        if fields == '' or values == '':
            return 0
        sql = str(" insert into " + self.__name + " ( " + fields + ") values ( " + values + " ) ")
        pk = 0
        try:
            self.__connect()
            self.cursor.execute(sql)
            pk = self.__conn.insert_id()
            self.__close()
        except Exception as e:
            self.__conn.rollback()
        return pk

    def setOption(self, key, val):
        return self.update({key: val})

    def insertAll(self, datas):
        if typeof(datas) != 'list':
            return None
        count = 0
        for data in datas:
            count += self.insert(data)
        return count

    def delete(self):
        i = 0
        sql = ''
        if len(self.__map) > 0:
            sql += ' where 1=1 '
            for item in self.__map:
                if typeof(item) == 'str':
                    sql += ' and ( ' + item + ' ) '
                elif typeof(item) == 'dict':
                    sql += ' and ( ' + item.get('key') + ' ' + item.get('type') + ' ' + str(item.get('val')) + ' ) '
        else:
            raise Exception("禁止不使用 where 删除数据")
        sql = str("delete from " + self.__name + sql)
        return self.__edit(sql)

    def setInc(self, key, step=1):
        return self.update({key: key + '+' + str(step)})

    def setDec(self, key, step=1):
        return self.update({key: key + '-' + str(step)})

    def query(self, sql):
        return self.__edit(sql)

    def __showColumn(self, is_str=False, table=None):
        if table is None:
            table = self.__name
        list_data = None
        sql = "SHOW FULL COLUMNS FROM " + table
        try:
            self.__connect()
            self.cursor.execute(sql)
            list_data = self.cursor.fetchall()
            self.__close()
        except Exception as e:
            logger.exception(e)
            raise e
        if is_str:
            return ','.join(list([item[0] for item in list_data]))
        return list([{'field': item[0], 'type': item[1], 'key': item[4]} for item in list_data])

    def __getPk(self):
        fields = self.__showColumn()
        for field in fields:
            if field['key'] == 'PRI':
                return field['field']
        return None

    def __edit(self, sql):
        count = 0
        try:
            self.__connect()
            count = self.cursor.execute(sql)
            self.__conn.commit()
            self.__close()
        except Exception as e:
            self.__conn.rollback()
            logger.error(sql)
            logger.exception(e)
            raise e
        return count

    def clear(self):
        self.__map = []
        self.__name = ''
        self.__column = '*'
        self.__alias = ''
        self.__join = []
        self.__having = []
        self.__distinct = False
        self.__option = {}





class ResponseException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class ResponseRenderException(ResponseException):

    def __init__(self, tpl_name, data):
        Exception.__init__(self, tpl_name)
        self.tpl_name = tpl_name
        self.data = data


class ResponseTextException(ResponseException):

    def __init__(self, text):
        Exception.__init__(self, ResponseTextException.__name__)
        self.text = text


class ResponseNotFoundException(ResponseTextException):
    pass



"""
    模型基类
"""

import logging, copy
from .query import QuerySet
from .contain.columns import ColumnCmp, iscolumn, ModelAttr
from .contain.func import *


logger = logging.getLogger(__name__)


class Model:
    """
    对基础的db方法进行ORM扩展和适配
    """
    __table__ = None        # 模型对象数据表表名
    __table_pk__ = None     # 主键
    __table_args__ = None   # 表加参数
    __table_fields__ = None # 模型的字段列表
    __attrs__ = None        # 模型显示的属性
    __relation__ = None     # 被关联信息
    __parent__ = []         # 父模型
    __origin__ = {}         # 原始数据（数据查询
    __data__ = {}           # 设置的数据
    __withs__ = []          # 预加载属性
    __query__ = QuerySet()

    def __init__(self, *args, **kwargs):
        self._is_new = False  # 是否是新增
        self.__modified_fields__ = []  # 模型数据是否修改，用于更新
        self.__relation__ = None    # 关联模型属性
        self.__withs__ = []         # 预加载属性
        self._load_field()          # 初始化字段信息
        # 对象数据初始化
        for key in self.__attrs__:
            if key in kwargs:
                setattr(self, key, kwargs.pop(key))
            else:
                cls_attr = object.__getattribute__(self.__class__, key)
                setattr(self, key, copy.copy(cls_attr))
        # 查询器
        # self.__query = QuerySet().table(self.__table__)   # 基础查询构造器
        self.__query__ = QuerySet().table(self.__table__)

    def _load_field(self):
        self.__class__._get_cls_fields()

    @property
    def name(self):
        return self.__class__.__name__

    @classmethod
    def _get_cls_fields(cls):
        if cls.__table_fields__ is not None:
            return
        cls.__table_fields__ = {}
        cls.__attrs__ = {}
        cls_dict = dict(cls.__dict__)
        for name in cls_dict:
            if str(name).find("__") == 0:
                continue
            # obj = getattr(cls, name)
            obj = cls.__getattribute__(cls, name, True)
            if isfunction(obj):
                continue
            if iscolumn(obj.__class__):
                # logger.debug("加载模型字段 cls={}, name={}, obj={}".format(cls, name, obj))
                obj.name = name
                if obj.primary_key:
                    # logger.debug("设置主键 {}={}:{}".format(name, type(obj), obj))
                    cls.__table_pk__ = obj
                cls.__table_fields__[name] = obj
                cls.__attrs__[name] = obj

            if issubclass(obj.__class__, RelationModel):
                obj.name = name
                cls.__attrs__[name] = obj

        # logger.debug("加载模型字段 cls={}, primary_key={}".format(cls, cls.__table_pk__))

    def where(self, *args, **kwargs):
        args = list(args)
        where = {'key': None, 'type': '=', 'val': None}
        if len(args) > 0 and isinstance(args[0], ColumnCmp):
            where = args[0].__dict__()
        elif len(args) == 1 and isinstance(args[0], dict):
            self.__query__ = self.__query__.where(args[0])
            return self
        else:
            where['key'] = args[0]
            if len(args) == 2:
                where['val'] = args[1]
            elif len(args) == 3:
                where['type'] = args[1]
                where['val'] = args[2]

        if len(self.__withs__) > 0 and len(args) >= 2 and where['key'].find(".") == -1:
            where['key'] = self.__table__ + "." + where['key']

        self.__query__ = self.__query__.where(where['key'], where['type'], where['val'])
        return self

    def get(self, *args, **kwargs):
        if len(args) > 0 or len(kwargs) > 0:
            self.where(*args, **kwargs)
        return self.find()

    def find(self):
        # 预加载处理（预载入模式开启后，数据返回将进行关联数据填充
        if not_empty(self.__withs__):
            # 加载本表字段
            self.__query__.field(True, False, self.__table__)
            # 处理关联表
            for it in self.__withs__:
                it.load_model_cls(self)
                it.eagerly(self.__query__)

        # 执行查询
        row = self.__query__.find()
        if row is None:
            return None

        # 数据装载
        obj_data = {}
        for key in self.__table_fields__:
            if key in row:
                obj_data[key] = row[key]

        # obj = object.__new__(self.__class__)  # 另外一种模型new方法
        # obj.__init__()
        obj = self.__class__(**obj_data)
        obj.__relation__ = self.__relation__
        obj.__modified_fields__ = []

        # 预载入结果
        if not_empty(self.__withs__):
            obj.eagerly_result(row, self.__withs__)

        return obj

    def all(self):
        return self.select()

    def select(self):
        # 预加载处理（预载入模式开启后，数据返回将进行关联数据填充
        if not_empty(self.__withs__):
            # 加载本表字段
            self.__query__.field(True, False, self.__table__)
            # 处理关联表
            for it in self.__withs__:
                it.load_model_cls(self)
                it.eagerly(self.__query__)

        # 执行查询
        rows = self.__query__.select()
        if empty(rows):
            return []

        # 数据装载
        ret = []
        for row in rows:
            obj_data = {}
            for key in self.__table_fields__:
                if key in row:
                    obj_data[key] = row[key]

            obj = self.__class__(**row)
            obj.__relation__ = self.__relation__
            obj.__modified_fields__ = []
            ret.append(obj)
            # 预载入结果
            if not_empty(self.__withs__):
                obj.eagerly_result(row, self.__withs__)

        # logger.debug("ret={}".format(ret))
        return ret

    def __query___soft_delete(self):
        """追加软删除列查询条件"""
        # self.__query__.where({})
        return self

    def save(self, data=None):
        # 赋值
        self.__load_data(data)

        if self.__is_modified__() is False:
            return

        """
        获取修改列的值，获取到的值经过__getattribute__处理会是直接的基本数据类型
        """
        update_data = {key: getattr(self, key) for key in self.__modified_fields__}

        # 新增还是更新
        pk = self.__class__.__table_pk__
        if pk is None or pk.name is None:
            raise Exception("模型ORM不支持没有主键的model.save操作")

        pk_val = getattr(self, pk.name)
        if pk_val is None:
            # 新增还涉及默认数据的列：default、insert_default、update_default
            for name, col in self.__class__.__table_fields__.items():
                if col.has_insert_default() and (name not in update_data or update_data[name] is None):
                    update_data[name] = col.__get_insert_default__()

            update_data = self.__data_sort(update_data)
            pk_val = self.__query__.insert_get_id(update_data)      # 执行插入
            # logger.debug("主键值={}".format(pk_val))
            setattr(self, pk.name, pk_val)

        else:
            # 更新涉及的数据列：update_default
            for name, col in self.__class__.__table_fields__.items():
                if col.has_update_default() and (name not in update_data or update_data[name] is None):
                    update_data[name] = col.__get_update_default__()

            self.where(pk == pk_val)
            update_data = self.__data_sort(update_data)
            self.__query__.update(update_data)

        self.__modified_fields__ = []
        return 1

    def remove(self):
        if len(self.__query__.get_map()) == 0:
            pk, pk_val = self.__get_pk_val__()
            self.where(pk == pk_val)
        return self.__query__.delete()

    def __get_pk_val__(self):
        pk = self.__class__.__table_pk__
        return pk, None if pk is None else getattr(self, pk.name)

    def __data_sort(self, data):
        """数据根据字段顺序排序"""
        data2 = {}
        for col_name, col in self.__class__.__table_fields__.items():
            for it, val in data.items():
                if col_name == it:
                    data2[col_name] = val
        return data2

    """
        批量方法
    """
    def update(self, data):
        # 是否未加载就保存，那就先查询
        if self.__is_load() and len(self.__query__.get_map()) > 0:
            self.__query__.update(data)

    def delete(self):
        # 是否未加载就保存，那就先查询
        if self.__is_load() and len(self.__query__.get_map()) > 0:
            self.__query__.delete()


    """
        属性控制
    """
    def __is_modified__(self):
        return len(self.__modified_fields__) > 0

    def __is_load(self):
        return self.__get_pk_val__() is not None

    def __load_data(self, data):
        if data is not None:
            for key in data:
                if key in self.__attrs__:
                    self.__setattr__(key, data[key])

    def __setattr__(self, name: str, value, attr_direct=False):
        # logger.debug("__setattr__ {}:{}".format(id(self), key))
        if name.find("_") == 0:
            return object.__setattr__(self, name, value)

        if attr_direct is False:        # 非直接操作，就加载下数据
            self.__relation_load_data()
            self.__modified_fields__.append(name)

        object.__setattr__(self, name, value)
        # logger.debug("__setattr__ ok {}:{}".format(id(self), key))

    def __getattr__(self, name):
        # 在model上未找到的属性，到
        # print("在model上未找到的属性", name)
        if hasattr(self.__query__, name):

            def proxy_method(*args, **kwargs):
                method = getattr(self.__query__, name)
                ret = method(*args, **kwargs)
                # print("代理方法-执行结果", method, ret)
                if isinstance(ret, QuerySet):
                    self.__query__ = ret
                    return self
                return ret

            return proxy_method

        # raise Exception("未在模型对象{}上找到名为'{}'的属性或方法".format(self.name, name))
        return None

    def __getattribute__(self, attr_name, attr_direct=False):
        attr = object.__getattribute__(self, attr_name)
        if attr_name.find("_") == 0 or attr_direct:     # 自带属性直接返回
            return attr
        if isinstance(attr, types.MethodType):              # 方法直接返回
            return attr

        # logger.debug("{}=>{}:{}".format(property_name, type(attr), attr))

        # 下级属性是关联模型
        self.__relation_load_data()
        attr = object.__getattribute__(self, attr_name)

        if iscolumn(attr.__class__):
            return attr.value

        if issubclass(attr.__class__, RelationModel):
            # logger.debug("关联模型属性")
            attr.load_model_cls(self)
            attr.bind_parent_where()
            object.__setattr__(self, attr_name, attr.model)
            # logger.debug("关联模型属性 返回新的模型={}".format(id(attr.model)))
            # logger.debug("关联模型属性 返回新的模型={}".format(type(attr.model)))
            # logger.debug("关联模型属性 返回新的模型={}".format(attr.model.__relation__))
            return attr.model

        return attr

    def __relation_load_data(self):
        __relation__ = object.__getattribute__(self, '__relation__')
        if __relation__ is not None and __relation__.is_load_data is False:
            if isinstance(__relation__, OneToOne):     # 只懒加载一对一
                data = __relation__.__load_data__()
                if isinstance(data, self.__class__):
                    for key in self.__attrs__:
                        self.__setattr__(key, getattr(data, key), True)
                    self.__modified_fields__ = []       # 可能由于顺序问题
                    # setattr(__relation__.parent, __relation__.name, self)       # 保持本对象
                elif isinstance(data, list):
                    setattr(__relation__.parent, __relation__.name, data)

    def __str__(self):
        pk, pk_val = self.__get_pk_val__()
        return "{} object({})=>{}".format(self.__class__.__name__, pk_val, self.__dict__())

    # def __repr__(self):
    #     return self.__str__()

    def __dict__(self):
        data = {}
        withs_name = [it.name for it in self.__withs__]
        for k, c in self.__attrs__.items():
            if len(withs_name) > 0 and isinstance(c, RelationModel):
                if k in withs_name:
                    data[k] = getattr(self, k)
            else:
                data[k] = getattr(self, k)

        return data

    def to_dict(self):
        return self.__dict__()

    """
        模型预加载
    """

    def withs(self, models):
        """预加载关联模型数据"""
        if empty(models):
            return self

        if isinstance(models, str):
            models = models.split(',')

        withs = models if isinstance(models, list) else models
        for it in withs:
            if isinstance(it, OneToOne):    # 只支持一对一的预载入
                self.__withs__.append(it)

        return self

    def eagerly_result(self, result, withs):
        """预载入关联查询 装载数据"""
        self.__withs__ = withs
        # 关联数据载入
        for it in self.__withs__:
            item = object.__getattribute__(self, it.name)
            item.load_model_cls(self)
            item.withs_result(result)


"""
    模型关联
"""


class RelationModel:
    type_ = None

    def __init__(self, *args, **kwargs):
        self.relation_name = args[0]    # type:quickpython.database.model   # 关联模型名称，该模型应该和对应模型在同一目录，否则应该写全路径，支出表名查找模型
        self.relation_key = kwargs.pop('relation_key', None)    # 关联模型字段
        self.local_key = kwargs.pop('local_key', None)          # 当前模型字段
        self.eagerly_type = kwargs.pop('eagerly_type', 0)       # 预载入方式 0 -JOIN 1 -IN
        self.model = None
        self.model_cls = None
        self.is_load_data = False
        self.name = None        # 父模型关联字段名称、本类对应的属性名称
        self.parent = None    # 父模型对象
        self.local_key_val = None    # 父模型关联字段名称、本类对应的属性名称
        # 基础查询器
        self.__query = None     # QuerySet().table()

    def load_model_cls(self, parent):
        """获取对应模型并进行查询-到调用此方法说明是单独调用的"""
        if self.model is not None:
            return self.model
        # 父模型的关联字段
        if self.local_key is None:
            self.local_key = "{}_id".format(self.name)

        self.parent = parent
        if hasattr(parent, self.local_key) is False:
            raise Exception("未找到父模型的关联字段 {}".format(self.local_key))

        # 获取关联模型类
        cls_path = "{}.{}".format(parent.__class__.__module__, self.relation_name)
        self.model_cls = load_cls(cls_path)
        if self.model_cls.__table_pk__ is None:
            self.model_cls()

        if self.relation_key is None:
            self.relation_key = self.model_cls.__table_pk__.name

        return self.model

    def bind_parent_where(self):
        self.local_key_val = getattr(self.parent, self.local_key)
        if self.local_key_val is None:
            logger.warning("父模型关联属性值为空 {}".format(self.local_key))
            return None

        self.model = self.model_cls()
        self.model.__relation__ = self
        self.model.where(self.relation_key, '=', self.local_key_val)

    def __load_data__(self):
        if self.is_load_data:
            return self.model

        self.is_load_data = True
        self.bind_parent_where()

        if isinstance(self, OneToOne):
            self.model = self.model.find()
            # logger.debug('self.model.__modified_fields__={}'.format(self.model.__modified_fields__))

        elif isinstance(self, OneToMany):
            self.model = self.model.select()

        return self.model

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def eagerly(self, query):
        """预载入关联查询（JOIN方式）"""
        relation = self.name
        # 处理父查询本字段，排除掉
        # 再父查询追加本模型的预载入查询
        join_table = self.model_cls.__table__
        join_alias = relation
        local_table = self.parent.__table__
        # field
        query.field(True, False, join_table, join_alias, relation + "__")
        # join
        join_on = "{}.{}={}.{}".format(local_table, self.local_key, join_alias, self.relation_key)
        query.join({join_table: join_alias}, join_on, 'LEFT')

    def withs_result_set(self, result_set):
        """预载入关联查询（数据集、多个）"""
        relation = self.name
        if self.eagerly_type == 0:
            for idx, result in result_set:
                self.match(relation, result)

    def withs_result(self, result):
        """预载入关联查询（数据、单个）"""
        relation = self.name
        if self.eagerly_type == 0:
            self.match(relation, result)

    def match(self, relation, result):
        """一对一 关联模型预查询拼装"""
        data = {relation: {}}
        for key, val in result.items():  # type:str, dict
            if key.find("__") > -1:
                name, attr = key.split("__")
                if name == relation:
                    data[name][attr] = val

        if not_empty(data[relation]):
            relation_model = self.model_cls(**data[relation])
            relation_model.__modified_fields__ = []
            relation_model.parent = self.parent     # 假设
        else:
            relation_model = None

        self.bind_parent_where()
        self.is_load_data = True
        self.model = relation_model
        self.parent.__setattr__(self.name, self.model, True)
        # logger.debug("数据装载完成：{}:{}".format(id(self.model), self.model))
        # parent_relation = getattr(self.parent, self.name)
        # logger.debug("数据装载完成-父属性值：{}:{}".format(id(parent_relation), parent_relation))


class OneToOne(RelationModel):
    type_ = "OneToOne"


class OneToMany(RelationModel):
    type_ = "OneToMany"


class HasManyThrough(RelationModel):
    type_ = "HasManyThrough"



"""
    引擎基类
"""
from ..contain.func import *


class Builder:

    def __init__(self, query):
        self.query = query

    def parse_field(self, fields):
        pass

    def build_where(self, where):
        if empty(where):
            where = []

        where_str = ""
        for key, val in where:
            str_arr = []
            for field, value in val:
                pass

    def parse_join(self, join):
        pass




"""
    python-quick.env 1.0
    环境变量管理器
"""
import os, logging


class Env:

    def __init__(self, env_path=".env", encoding="utf-8"):
        self.env_path = env_path
        self.encoding = encoding
        self._DATA = {}
        self.load_env(env_path)

    def load_env(self, env_path):
        """加载本地.env文件"""
        if os.path.exists(env_path) is False:
            raise "env file '{}' not found".format(env_path)
            # return logging.warning("env file '{}' not found".format(env_path))

        self.encoding = "utf-8" if self.encoding is None else self.encoding

        with open(env_path, encoding=self.encoding) as fp:
            lines = fp.readlines()
            group_name = None
            for line in lines:
                if line[-1] == "\n":
                    line = line[:-1]
                if line == '' or line[0] == '#':
                    continue
                # 组名
                if line[0] == '[' and line[-1] == ']':
                    group_name = line[1:-1]
                    self._DATA[group_name] = {}
                    continue
                # 数据
                ei = line.find("=")
                if ei == -1:
                    continue

                key, val = line[:ei].strip(), line[ei+1:].strip()
                # 数据实例化
                if len(val) == 0: val = ''
                elif val.lower() == "true": val = True
                elif val.lower() == "false": val = False

                if group_name is not None:
                    self._DATA[group_name][key] = val
                    self._write_os_env(group_name + '.' + key, val)
                else:
                    self._DATA[key] = val
                    self._write_os_env(key, val)

    def _write_os_env(self, k, v):
        if isinstance(v, str):
            os.environ.setdefault(k, v)
        else:
            os.environ.setdefault(k, '')

    def get(self, key:str, def_val=None):
        if key.find(".") > -1:
            key1, key2 = tuple(key.split("."))
            if key1 not in self._DATA:
                return def_val
            group = self._DATA.get(key1)
            return group.get(key2) if key2 in group else def_val
        else:
            return self._DATA.get(key) if key in self._DATA else def_val

    def set(self, key, val):
        if key.find(".") > -1:
            key1, key2 = tuple(key.split("."))
            if key1 not in self._DATA:
                self._DATA[key1] = {}
            self._DATA[key1][key2] = val
        else:
            self._DATA[key] = val

        return val

    @classmethod
    def __get_global__(cls):
        env = globals().get("__quickpython.component.env")
        if env is None:
            env = Env()
            globals()["__quickpython.component.env"] = env

        return env


env = Env.__get_global__()


if __name__ == "__main__":
    print("env", env)
    print("env", env.get("app"))
    print("env", env.get("app.debug"))
    print("env", env.get("database"))
    # 设置
    print("set", env.set("app.asd", "asd"))
    print("set", env.set("app.asd", True))
    print("set", env.get("app"))




#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

setup(
    name="quickpython-mvc",
    version="0.1.3",
    keywords=["quickpython-mvc", "quickpython", "quick", "WEB", "database", "MVC"],  # 关键字
    description="Python rapid development framework .",
    long_description="Python rapid development framework ",
    license="MIT Licence",

    url="https://github.com/quickpython",
    author="lo106258",
    author_email="lo106258@gmail.com",

    packages=find_packages(),
    include_package_data=True,
    scripts=['./boot.py', './config.py', './README.md', './requirements.txt'],
    platforms="any",
    install_requires=["tornado", "pymysql"],
    extras_require={':python_version <= "3.6"': ['enum34', 'future']}
)



"""
model demo
"""
import logging
from app.common.models import *
from libs.utils import Utils

logger = logging.getLogger(__name__)


class DemoModel:

    def call(self):
        """"""
        # self.base()
        self.join()

    def base(self):
        # self.save()
        # self.remove()
        # self.save_update()
        self.get()
        # self.field_use()

    def join(self):
        self.join_1()
        # self.join_2()

    def save(self):
        """新增"""
        mtime = Utils.mtime()
        UserModel(**{'username': "test1"}).save()
        UserModel(**{'username': "test2"}).save()
        UserModel(**{'username': "test3"}).save()
        UserModel(**{'username': "test4"}).save()
        UserModel(**{'username': "test5"}).save()
        logger.info("新增耗时 {}ms".format(Utils.mtime() - mtime))

    def remove(self):
        """删除"""
        user_2 = UserModel().where(UserModel.username == "test2").find()
        if user_2 is None:
            raise Exception("未找到nickname = test2 的数据")
        else:
            user_2.remove()
        logger.info("删除成功")

    def save_update(self):
        """更新"""
        user_3 = UserModel().where(UserModel.username == "test3").find()
        if user_3 is None:
            raise Exception("未找到nickname = test3 的数据")
        else:
            user_3.nickname = user_3.username
            user_3.save()
        logger.info("更新成功")

    def get(self):
        user_4 = UserModel().where(UserModel.username == "test4").get()
        if user_4 is None:
            raise Exception("未找到nickname = test4 的数据")
        logger.info("user_4={}".format(user_4.id))
        logger.info("user_4={}".format(user_4))
        # 多条
        # user_rows = UserModel().where(UserModel.username == "test4").select()
        # logger.info("user_rows={}:{}".format(len(user_rows), user_rows))
        # logger.info("查找成功")

    def field_use(self):
        """字段使用和参与计算"""
        user_5 = UserModel().where(UserModel.username == "test5").find()
        if user_5 is None:
            logger.info("未找到的数据")

        logger.info("修改前 user.school_id={}".format(user_5.school_id))
        user_5.school_id = user_5.school_id + 1
        logger.info("修改后 user.school_id={}".format(user_5.school_id))
        user_5.save()
        logger.info("保存后 user_5={}".format(user_5))

    def join_1(self):
        """模型预载入查询单条 一对一（left join）"""

        # 预加载
        # user_6 = UserModel().withs([UserModel.device])\
        #     .where("user.id", 40830)\
        #     .find()
        # Utils.print_dicts(user_6)
        # Utils.print_dicts(user_6.device)

        # 懒加载
        # user_6 = UserModel().where("user.id", 40830).get()
        # Utils.print_dicts(user_6)
        # print("user_6.device", id(user_6.device), user_6.device)
        # print("user_6.device", id(user_6.device), user_6.device)
        # user_6.device.key = Utils.datetime_now()
        # user_6.device.save()
        # print("user_6.device", id(user_6.device), user_6.device)

        # print("user_6.device", user_6.device.__query__.get_map())
        # user_6.device.save()
        # logger.debug("user_6.device={}".format(user_6.device))
        # logger.debug("user_6.device={}".format(user_6.device.where('platform', 2).get()))
        # logger.debug("user_6.school={}".format(user_6.school))

        # logger.debug("user_6.money_log={}".format(user_6.money_log))
        # money_log = user_6.money_log.order('id', 'DESC').limit(10).select()
        # logger.debug("user_6.money_log={}:{}".format(len(money_log), money_log))
        # Utils.print_dicts(user_6)
        # Utils.print_dicts(user_6.device)

    def join_2(self):
        """模型预载入查询 多个"""
        user_7 = UserModel().withs([UserModel.device, UserModel.school])\
            .where(UserModel.id == 40850)\
            .where(UserModel.branch_id == 1062)\
            .select()
        logger.debug("user_7=\n{}".format(Utils.dict_to_str(user_7)))


if __name__ == '__main__':
    try:
        DemoModel().call()
    except:
        Utils.print_exc()



"""
    Session管理器
    python-quick.session 1.0
    PS: 适配tornado

"""
import os, time
from hashlib import sha1
from quickpython.component.env import env


class Session2:

    def __init__(self, sess_id):
        self.id = sess_id


class Session:

    name = env.get('session.name', 'qp_session')
    expire = env.get('session.expire', 60)

    _CACHE = {
        # session_id: {}
    }

    def __init__(self, handler, session_name=None):
        self.handler = handler
        self.name = self.name if session_name is None else session_name
        self.sess_id = self.handler.get_cookie(self.name)
        if self.sess_id is None:
            self.sess_id = self._gen_session_id()
            self._CACHE[self.sess_id] = {}
            self.handler.set_cookie(self.name, self.sess_id, max_age=self.expire)

    @classmethod
    def _gen_session_id(cls):
        return sha1(bytes('%s%s' % (time.time(), os.urandom(32)), encoding='utf-8')).hexdigest()

    def __getitem__(self, item):
        return self._CACHE[self.sess_id].get(item)

    def __setitem__(self, key, value):
        self._CACHE[self.sess_id][key] = value

    def __delitem__(self, key):
        if key in self._CACHE[self.sess_id]:
            del self._CACHE[self.sess_id][key]

    def get(self, key, def_val=None):
        if self.sess_id in self._CACHE:
            return self._CACHE[self.sess_id].get(key, def_val)
        return def_val

    def set(self, key, val):
        if self.sess_id not in self._CACHE:
            self._CACHE[self.sess_id] = {}
        self._CACHE[self.sess_id][key] = val

    def delete(self):
        """从大字典删除session_id"""
        del self._CACHE[self.sess_id]




import os, logging, urllib3, threading, datetime
from multiprocessing import cpu_count
from logging.handlers import TimedRotatingFileHandler
from quickpython.component.env import env

logger = logging.getLogger(__name__)
# 屏蔽ssl警告
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Config:

    APP = 'app'     # 应用目录
    MODE = None     # 运行模式：web、cmd
    DEBUG = True
    ROOT_PATH = str(os.getcwd()).replace('\\', '/')

    #
    MODE_WEB = "WEB"
    MODE_CMD = "CMD"

    # 目录
    CACHE_PATH = ROOT_PATH + '/cache'       # cache目录
    DATA_PATH = ROOT_PATH + '/data'         # 数据目录
    LOGS_PATH = ROOT_PATH + '/cache/logs'   # 日志目录
    PUBLIC_PATH = ROOT_PATH + '/public'     # 资源目录
    UPLOADS_PATH = ROOT_PATH + '/public/uploads'    # 上传目录
    DIRS = [CACHE_PATH, LOGS_PATH, DATA_PATH, UPLOADS_PATH]

    @classmethod
    def init(cls, mode):
        cls.MODE = str(mode).upper()
        cls.logging_init()
        for dir_it in cls.DIRS:
            if os.path.exists(dir_it) is False:
                logger.info('创建目录：{}'.format(dir_it))
                os.makedirs(dir_it)

    @staticmethod
    def is_debug():
        return env.get("app.debug", False)

    @classmethod
    def logging_init(cls):
        """初始化日志模块"""
        logging.getLogger("matplotlib").setLevel(logging.WARNING)
        logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
        # 控制台
        console_handler = logging.StreamHandler()
        # 文件
        encoding = env.get("encoding", "UTF-8")
        file_path = '{}/{}'.format(cls.LOGS_PATH, cls.get_log_file_name())
        file_handler = TimedRotatingFileHandler(file_path, when='D', backupCount=33, encoding=encoding)
        file_handler.setLevel(logging.DEBUG)
        # 全局
        formatter_str = '%(asctime)s %(levelname)s [%(filename)s:%(funcName)s:%(lineno)d]\t%(message)s'
        level = logging.DEBUG if cls.is_debug() else logging.INFO
        logging.basicConfig(format=formatter_str, level=level, handlers=[console_handler, file_handler])

    # env环境变量
    @staticmethod
    def env_get(key, default=None):
        val = env.get(key)
        if val == "true":
            val = True
        elif val == "false":
            val = False

        if val is None and default is not None:
            return default
        return val

    @staticmethod
    def env_get_int(key, default=None):
        try:
            val = env.get(key)
            return default if val is None else int(val)
        except:
            logger.error("env:{} get error".format(key))
            return default

    @classmethod
    def get_logger(cls, name):
        return logging.getLogger(name)

    @classmethod
    def get_log_file_name(cls):
        if cls.MODE == cls.MODE_CMD:
            return "{}_{}.log".format(cls.MODE_CMD.lower(), datetime.datetime.now().strftime("%Y-%m-%d"))
        return "{}.log".format(cls.MODE_CMD.lower())

    @classmethod
    def web_pro_count(cls, num=None):
        return cpu_count() if num is None else num

    @classmethod
    def web_thr_count(cls, num=None):
        # thr_num = cpu_count() * 2 if num is None else num
        # thr_num = 4 if cls.is_debug() else thr_num
        return 4




import sys, signal, logging, datetime
import tornado.ioloop
import tornado.web
from quickpython.config import Config
from .settings import SETTINGS, ROUTES


class Core:
    log = logging.getLogger(__name__)

    @classmethod
    def signal_init(cls):
        """监听退出"""

        def signin_exit(signum, frame):
            cls.log.info("进程终止信号")
            tornado.ioloop.IOLoop.instance().stop()
            cls.log.info("进程终止完成")

        # signal.signal(signal.SIGQUIT, signin_exit)
        signal.signal(signal.SIGTERM, signin_exit)
        signal.signal(signal.SIGINT, signin_exit)

        # tornado专属停止方式
        def set_ping(ioloop, timeout):
            ioloop.add_timeout(timeout, lambda: set_ping(ioloop, timeout))

        set_ping(tornado.ioloop.IOLoop.instance(), datetime.timedelta(seconds=1))

    @classmethod
    def init(cls, mode):
        Config.init(mode)
        cls.signal_init()
        logging.getLogger("tornado.access").setLevel(logging.ERROR)

    @classmethod
    def cmd(cls):
        cls.init(Config.MODE_CMD)
        cls.log.info("##########   Command mode   ##########")

    @classmethod
    def start(cls):
        """启动web环境"""
        cls.init(Config.MODE_WEB)
        # 配置
        application = tornado.web.Application(ROUTES, **{
            'template_path': SETTINGS['template_path'],
            'ui_methods': SETTINGS['ui_methods'],
        })  # , debug=SETTINGS['debug'])
        # 启动web
        cls.log.info("WEB start port " + str(SETTINGS['port']))
        server = cls.server = tornado.web.HTTPServer(application, decompress_request=True)
        if sys.platform == "win32":
            server.listen(SETTINGS['port'])
        else:
            server.bind(SETTINGS['port'])
            server.start(Config.web_pro_count(SETTINGS['pro_thr_num']))

        cls.log.debug("web start complete.")
        tornado.ioloop.IOLoop.instance().start()
        # 关闭
        # tornado.ioloop.IOLoop.current().stop()


if __name__ == "__main__":
    Core.start()



"""
model demo
"""
import logging
from libs.utils import Utils
from quickpython.database.contain.columns import *
from quickpython.database.connector.connection import Connection
from quickpython.component.env import env

logger = logging.getLogger(__name__)


class GenerateDemo:

    def call(self):
        """"""
        self.base()

    def base(self):
        """
        数据库表模型生成器
        :return:
        """
        conn = Connection().connect()
        # 获取全部表
        count, ret, field_info = conn.execute_all("SHOW TABLES")
        prefix = conn.get_config('prefix')
        table_dict = {it[0]: {} for it in ret}
        # 获取表的字段属性
        for table in table_dict:
            sql = "SHOW FULL COLUMNS FROM `{}`".format(table)
            ret2 = conn.execute_all(sql)[1]
            for it in ret2:
                field_name = it[0]
                type_ = str(it[1]).split("(")[0].lower()
                type_cls = field_map[type_] if type_ in field_map else field_map['varchar']
                table_dict[table][field_name] = {
                    'name': field_name,
                    'type': type_,
                    'type_cls': type_cls,
                    'primary_key': it[4] == "PRI",
                    'comment': it[8],
                }

        # 格式化输出
        ret = []
        for it in table_dict:
            ret.append(self.format_table(it, table_dict[it], prefix))

        print("\n\n\n".join(ret))

    def format_table(self, table_name, fields, prefix):
        """"""
        field_arr = []
        for key in fields:
            it = fields[key]
            args = []
            if it['primary_key']:
                args.append("primary_key=True")
            if len(it['comment']) > 0:
                args.append("comment=\"{}\"".format(it['comment']))

            item = "{} = {}({})".format(it['name'], it['type_cls'].__name__, ", ".join(args))
            field_arr.append(item)

        model_name = table_name if prefix == '' else table_name.replace(prefix, '')
        model_name = Utils.str_to_hump(model_name)

        table_name = table_name if prefix == '' else table_name.replace(prefix, '')

        t_ = "    "
        text = """class {}Model(Model):\n{}__table__ = "{}"\n\n{}{}"""\
            .format(model_name, t_, table_name, t_, "\n    ".join(field_arr))
        return text


if __name__ == '__main__':
    try:
        GenerateDemo().call()
    except:
        Utils.print_exc()



import logging
from quickpython.component.session import Session
from quickpython.server.exception import *


class Controller:
    tpl_style = 'default'
    tpl_ext = 'html'

    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)
        self.pro_obj = None         # type:web.processor.ProcessorController
        self.params = {}            # 请求数据
        self.assign_data = {}       # 默认注入参数

    def initialize(self):
        pass

    def __initialize__(self, obj, module: str, controller: str, action: str):
        self.pro_obj = obj
        self.module = module
        self.controller = controller.replace('.', '/')
        self.action = action

    def __initialize_request__(self, path, params, request):
        self.path = path
        self.params = params
        self.headers = request.headers
        self.cookies = request.cookies
        self.session = Session(self)
        self.initialize()

    def render(self, *args, **kwargs):
        args = list(args)
        if args[0].find(self.tpl_ext) > -1:     # type:str
            args[0] = "{}/{}".format(self.controller, args[0])
        else:
            args[0] = "{}/{}.{}".format(self.controller, args[0], self.tpl_ext)
        args[0] = "{}/view/{}/{}".format(self.module, self.tpl_style, args[0])
        # return self.pro_obj.render(template_name=args[0], __config="", **{**self._assign_data, **kwargs})

        raise ResponseRenderException(args[0], {**{**self.assign_data, **kwargs}})
        # return args[0], {**{**self._assign_data, **kwargs}}

    def __getattr__(self, name):
        return getattr(self.pro_obj, name)

    def view(self, **kwargs):
        view_path = "{}.{}".format(self.action, self.tpl_ext)
        return self.render(view_path, **kwargs)
